# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from distutils.version import LooseVersion
import warnings
import logging
import datetime
import ast

import h5py
import numpy as np
import dask.array as da
from traits.api import Undefined
from hyperspy.misc.utils import ensure_unicode, multiply, get_object_package_info
from hyperspy.axes import AxesManager
from collections import namedtuple

_logger = logging.getLogger(__name__)


# Plugin characteristics
# ----------------------
format_name = 'HSPY'
description = \
    'The default file format for HyperSpy based on the HDF5 standard'
full_support = False
# Recognised file extension
file_extensions = ['hspy', 'hdf5']
default_extension = 0
# Writing capabilities
writes = True
version = "3.0"

# -----------------------
# File format description
# -----------------------
# The root must contain a group called Experiments
# The experiments group can contain any number of subgroups
# Each subgroup is an experiment or signal
# Each subgroup must contain at least one dataset called data
# The data is an array of arbitrary dimension
# In addition a number equal to the number of dimensions of the data
# dataset + 1 of empty groups called coordinates followed by a number
# must exists with the following attributes:
#    'name'
#    'offset'
#    'scale'
#    'units'
#    'size'
#    'index_in_array'
# The experiment group contains a number of attributes that will be
# directly assigned as class attributes of the Signal instance. In
# addition the experiment groups may contain 'original_metadata' and
# 'metadata'subgroup that will be
# assigned to the same name attributes of the Signal instance as a
# Dictionary Browsers
# The Experiments group can contain attributes that may be common to all
# the experiments and that will be accessible as attributes of the
# Experiments instance
#
# CHANGES
#
# v3.0
# - add Camera and Stage node
# - move tilt_stage to Stage.tilt_alpha
#
# v2.2
# - store more metadata as string: date, time, notes, authors and doi
# - store quantity for intensity axis
#
# v2.1
# - Store the navigate attribute.
# - record_by is stored only for backward compatibility but the axes navigate
#   attribute takes precendence over record_by for files with version >= 2.1
# v1.3
# ----
# - Added support for lists, tuples and binary strings

not_valid_format = 'The file is not a valid HyperSpy hdf5 file'

current_file_version = None  # Format version of the file being read
default_version = LooseVersion(version)


def get_hspy_format_version(f):
    if "file_format_version" in f.attrs:
        version = f.attrs["file_format_version"]
        if isinstance(version, bytes):
            version = version.decode()
        if isinstance(version, float):
            version = str(round(version, 2))
    elif "Experiments" in f:
        # Chances are that this is a HSpy hdf5 file version 1.0
        version = "1.0"
    elif "Analysis" in f:
        # Starting version 2.0 we have "Analysis" field as well
        version = "2.0"
    else:
        raise IOError(not_valid_format)
    return LooseVersion(version)


def file_reader(filename, backing_store=False,
                lazy=False, **kwds):
    """Read data from hdf5 files saved with the hyperspy hdf5 format specification

    Parameters
    ----------
    filename: str
    lazy: bool
        Load image lazily using dask
    **kwds, optional
    """
    mode = kwds.pop('mode', 'r')
    f = h5py.File(filename, mode=mode, **kwds)
    # Getting the format version here also checks if it is a valid HSpy
    # hdf5 file, so the following two lines must not be deleted or moved
    # elsewhere.
    global current_file_version
    current_file_version = get_hspy_format_version(f)
    global default_version
    if current_file_version > default_version:
        warnings.warn(
            "This file was written using a newer version of the "
            "HyperSpy hdf5 file format. I will attempt to load it, but, "
            "if I fail, it is likely that I will be more successful at "
            "this and other tasks if you upgrade me.")

    models_with_signals = []
    standalone_models = []
    if 'Analysis/models' in f:
        try:
            m_gr = f.require_group('Analysis/models')
            for model_name in m_gr:
                if '_signal' in m_gr[model_name].attrs:
                    key = m_gr[model_name].attrs['_signal']
                    # del m_gr[model_name].attrs['_signal']
                    res = hdfgroup2dict(
                        m_gr[model_name],
                        lazy=lazy)
                    del res['_signal']
                    models_with_signals.append((key, {model_name: res}))
                else:
                    standalone_models.append(
                        {model_name: hdfgroup2dict(
                            m_gr[model_name], lazy=lazy)})
        except TypeError:
            raise IOError(not_valid_format)

    experiments = []
    exp_dict_list = []
    if 'Experiments' in f:
        for ds in f['Experiments']:
            if isinstance(f['Experiments'][ds], h5py.Group):
                if 'data' in f['Experiments'][ds]:
                    experiments.append(ds)
        # Parse the file
        for experiment in experiments:
            exg = f['Experiments'][experiment]
            exp = hdfgroup2signaldict(exg, lazy)
            # assign correct models, if found:
            _tmp = {}
            for (key, _dict) in reversed(models_with_signals):
                if key == exg.name:
                    _tmp.update(_dict)
                    models_with_signals.remove((key, _dict))
            exp['models'] = _tmp

            exp_dict_list.append(exp)

    for _, m in models_with_signals:
        standalone_models.append(m)

    exp_dict_list.extend(standalone_models)
    if not len(exp_dict_list):
        raise IOError('This is not a valid HyperSpy HDF5 file. '
                      'You can still load the data using a hdf5 reader, '
                      'e.g. h5py, and manually create a Signal. '
                      'Please, refer to the User Guide for details')
    if not lazy:
        f.close()
    return exp_dict_list


def hdfgroup2signaldict(group, lazy=False):
    global current_file_version
    global default_version
    if current_file_version < LooseVersion("1.2"):
        metadata = "mapped_parameters"
        original_metadata = "original_parameters"
    else:
        metadata = "metadata"
        original_metadata = "original_metadata"

    exp = {'metadata': hdfgroup2dict(
        group[metadata], lazy=lazy),
        'original_metadata': hdfgroup2dict(
            group[original_metadata], lazy=lazy),
        'attributes': {}
    }
    if "package" in group.attrs:
        # HyperSpy version is >= 1.5
        exp["package"] = group.attrs["package"]
        exp["package_version"] = group.attrs["package_version"]
    else:
        # Prior to v1.4 we didn't store the package information. Since there
        # were already external package we cannot assume any package provider so
        # we leave this empty.
        exp["package"] = ""
        exp["package_version"] = ""

    data = group['data']
    if lazy:
        data = da.from_array(data, chunks=data.chunks)
        exp['attributes']['_lazy'] = True
    else:
        data = np.asanyarray(data)
    exp['data'] = data
    axes = []
    for i in range(len(exp['data'].shape)):
        try:
            axes.append(dict(group['axis-%i' % i].attrs))
            axis = axes[-1]
            for key, item in axis.items():
                if isinstance(item, np.bool_):
                    axis[key] = bool(item)
                else:
                    axis[key] = ensure_unicode(item)
        except KeyError:
            break
    if len(axes) != len(exp['data'].shape):  # broke from the previous loop
        try:
            axes = [i for k, i in sorted(iter(hdfgroup2dict(
                group['_list_' + str(len(exp['data'].shape)) + '_axes'],
                lazy=lazy).items()))]
        except KeyError:
            raise IOError(not_valid_format)
    exp['axes'] = axes
    if 'learning_results' in group.keys():
        exp['attributes']['learning_results'] = \
            hdfgroup2dict(
                group['learning_results'],
                lazy=lazy)
    if 'peak_learning_results' in group.keys():
        exp['attributes']['peak_learning_results'] = \
            hdfgroup2dict(
                group['peak_learning_results'],
                lazy=lazy)

    # If the title was not defined on writing the Experiment is
    # then called __unnamed__. The next "if" simply sets the title
    # back to the empty string
    if "General" in exp["metadata"] and "title" in exp["metadata"]["General"]:
        if '__unnamed__' == exp['metadata']['General']['title']:
            exp['metadata']["General"]['title'] = ''

    if current_file_version < LooseVersion("1.1"):
        # Load the decomposition results written with the old name,
        # mva_results
        if 'mva_results' in group.keys():
            exp['attributes']['learning_results'] = hdfgroup2dict(
                group['mva_results'], lazy=lazy)
        if 'peak_mva_results' in group.keys():
            exp['attributes']['peak_learning_results'] = hdfgroup2dict(
                group['peak_mva_results'], lazy=lazy)
        # Replace the old signal and name keys with their current names
        if 'signal' in exp['metadata']:
            if "Signal" not in exp["metadata"]:
                exp["metadata"]["Signal"] = {}
            exp['metadata']["Signal"]['signal_type'] = \
                exp['metadata']['signal']
            del exp['metadata']['signal']

        if 'name' in exp['metadata']:
            if "General" not in exp["metadata"]:
                exp["metadata"]["General"] = {}
            exp['metadata']['General']['title'] = \
                exp['metadata']['name']
            del exp['metadata']['name']

    if current_file_version < LooseVersion("1.2"):
        if '_internal_parameters' in exp['metadata']:
            exp['metadata']['_HyperSpy'] = \
                exp['metadata']['_internal_parameters']
            del exp['metadata']['_internal_parameters']
            if 'stacking_history' in exp['metadata']['_HyperSpy']:
                exp['metadata']['_HyperSpy']["Stacking_history"] = \
                    exp['metadata']['_HyperSpy']['stacking_history']
                del exp['metadata']['_HyperSpy']["stacking_history"]
            if 'folding' in exp['metadata']['_HyperSpy']:
                exp['metadata']['_HyperSpy']["Folding"] = \
                    exp['metadata']['_HyperSpy']['folding']
                del exp['metadata']['_HyperSpy']["folding"]
        if 'Variance_estimation' in exp['metadata']:
            if "Noise_properties" not in exp["metadata"]:
                exp["metadata"]["Noise_properties"] = {}
            exp['metadata']['Noise_properties']["Variance_linear_model"] = \
                exp['metadata']['Variance_estimation']
            del exp['metadata']['Variance_estimation']
        if "TEM" in exp["metadata"]:
            if "Acquisition_instrument" not in exp["metadata"]:
                exp["metadata"]["Acquisition_instrument"] = {}
            exp["metadata"]["Acquisition_instrument"]["TEM"] = \
                exp["metadata"]["TEM"]
            del exp["metadata"]["TEM"]
            tem = exp["metadata"]["Acquisition_instrument"]["TEM"]
            if "EELS" in tem:
                if "dwell_time" in tem:
                    tem["EELS"]["dwell_time"] = tem["dwell_time"]
                    del tem["dwell_time"]
                if "dwell_time_units" in tem:
                    tem["EELS"]["dwell_time_units"] = tem["dwell_time_units"]
                    del tem["dwell_time_units"]
                if "exposure" in tem:
                    tem["EELS"]["exposure"] = tem["exposure"]
                    del tem["exposure"]
                if "exposure_units" in tem:
                    tem["EELS"]["exposure_units"] = tem["exposure_units"]
                    del tem["exposure_units"]
                if "Detector" not in tem:
                    tem["Detector"] = {}
                tem["Detector"] = tem["EELS"]
                del tem["EELS"]
            if "EDS" in tem:
                if "Detector" not in tem:
                    tem["Detector"] = {}
                if "EDS" not in tem["Detector"]:
                    tem["Detector"]["EDS"] = {}
                tem["Detector"]["EDS"] = tem["EDS"]
                del tem["EDS"]
            del tem
        if "SEM" in exp["metadata"]:
            if "Acquisition_instrument" not in exp["metadata"]:
                exp["metadata"]["Acquisition_instrument"] = {}
            exp["metadata"]["Acquisition_instrument"]["SEM"] = \
                exp["metadata"]["SEM"]
            del exp["metadata"]["SEM"]
            sem = exp["metadata"]["Acquisition_instrument"]["SEM"]
            if "EDS" in sem:
                if "Detector" not in sem:
                    sem["Detector"] = {}
                if "EDS" not in sem["Detector"]:
                    sem["Detector"]["EDS"] = {}
                sem["Detector"]["EDS"] = sem["EDS"]
                del sem["EDS"]
            del sem

        if "Sample" in exp["metadata"] and "Xray_lines" in exp[
                "metadata"]["Sample"]:
            exp["metadata"]["Sample"]["xray_lines"] = exp[
                "metadata"]["Sample"]["Xray_lines"]
            del exp["metadata"]["Sample"]["Xray_lines"]

        for key in ["title", "date", "time", "original_filename"]:
            if key in exp["metadata"]:
                if "General" not in exp["metadata"]:
                    exp["metadata"]["General"] = {}
                exp["metadata"]["General"][key] = exp["metadata"][key]
                del exp["metadata"][key]
        for key in ["record_by", "signal_origin", "signal_type"]:
            if key in exp["metadata"]:
                if "Signal" not in exp["metadata"]:
                    exp["metadata"]["Signal"] = {}
                exp["metadata"]["Signal"][key] = exp["metadata"][key]
                del exp["metadata"][key]

    if current_file_version < LooseVersion("3.0"):
        if "Acquisition_instrument" in exp["metadata"]:
            # Move tilt_stage to Stage.tilt_alpha
            # Move exposure time to Detector.Camera.exposure_time
            if "TEM" in exp["metadata"]["Acquisition_instrument"]:
                tem = exp["metadata"]["Acquisition_instrument"]["TEM"]
                exposure = None
                if "tilt_stage" in tem:
                    tem["Stage"] = {"tilt_alpha": tem["tilt_stage"]}
                    del tem["tilt_stage"]
                if "exposure" in tem:
                    exposure = "exposure"
                # Digital_micrograph plugin was parsing to 'exposure_time'
                # instead of 'exposure': need this to be compatible with
                # previous behaviour
                if "exposure_time" in tem:
                    exposure = "exposure_time"
                if exposure is not None:
                    if "Detector" not in tem:
                        tem["Detector"] = {"Camera": {
                            "exposure": tem[exposure]}}
                    tem["Detector"]["Camera"] = {"exposure": tem[exposure]}
                    del tem[exposure]
            # Move tilt_stage to Stage.tilt_alpha
            if "SEM" in exp["metadata"]["Acquisition_instrument"]:
                sem = exp["metadata"]["Acquisition_instrument"]["SEM"]
                if "tilt_stage" in sem:
                    sem["Stage"] = {"tilt_alpha": sem["tilt_stage"]}
                    del sem["tilt_stage"]

    return exp


def dict2hdfgroup(dictionary, group, **kwds):
    "Recursive writer of dicts and signals"

    from hyperspy.misc.utils import DictionaryTreeBrowser
    from hyperspy.signal import BaseSignal

    def parse_structure(key, group, value, _type, **kwds):
        try:
            # Here we check if there are any signals in the container, as
            # casting a long list of signals to a numpy array takes a very long
            # time. So we check if there are any, and save numpy the trouble
            if np.any([isinstance(t, BaseSignal) for t in value]):
                tmp = np.array([[0]])
            else:
                tmp = np.array(value)
        except ValueError:
            tmp = np.array([[0]])
        if tmp.dtype == np.dtype('O') or tmp.ndim != 1:
            dict2hdfgroup(dict(zip(
                [str(i) for i in range(len(value))], value)),
                group.create_group(_type + str(len(value)) + '_' + key),
                **kwds)
        elif tmp.dtype.type is np.unicode_:
            if _type + key in group:
                del group[_type + key]
            group.create_dataset(_type + key,
                                 tmp.shape,
                                 dtype=h5py.special_dtype(vlen=str),
                                 **kwds)
            group[_type + key][:] = tmp[:]
        else:
            if _type + key in group:
                del group[_type + key]
            group.create_dataset(
                _type + key,
                data=tmp,
                **kwds)

    for key, value in dictionary.items():
        if isinstance(value, dict):
            dict2hdfgroup(value, group.create_group(key),
                          **kwds)
        elif isinstance(value, DictionaryTreeBrowser):
            dict2hdfgroup(value.as_dictionary(),
                          group.create_group(key),
                          **kwds)
        elif isinstance(value, BaseSignal):
            kn = key if key.startswith('_sig_') else '_sig_' + key
            write_signal(value, group.require_group(kn))
        elif isinstance(value, (np.ndarray, h5py.Dataset, da.Array)):
            overwrite_dataset(group, value, key, **kwds)
        elif value is None:
            group.attrs[key] = '_None_'
        elif isinstance(value, bytes):
            try:
                # binary string if has any null characters (otherwise not
                # supported by hdf5)
                value.index(b'\x00')
                group.attrs['_bs_' + key] = np.void(value)
            except ValueError:
                group.attrs[key] = value.decode()
        elif isinstance(value, str):
            group.attrs[key] = value
        elif isinstance(value, AxesManager):
            dict2hdfgroup(value.as_dictionary(),
                          group.create_group('_hspy_AxesManager_' + key),
                          **kwds)
        elif isinstance(value, list):
            if len(value):
                parse_structure(key, group, value, '_list_', **kwds)
            else:
                group.attrs['_list_empty_' + key] = '_None_'
        elif isinstance(value, tuple):
            if len(value):
                parse_structure(key, group, value, '_tuple_', **kwds)
            else:
                group.attrs['_tuple_empty_' + key] = '_None_'

        elif value is Undefined:
            continue
        else:
            try:
                group.attrs[key] = value
            except BaseException:
                _logger.exception(
                    "The hdf5 writer could not write the following "
                    "information in the file: %s : %s", key, value)


def get_signal_chunks(shape, dtype, signal_axes=None):
    """Function that claculates chunks for the signal, preferably at least one
    chunk per signal space.

    Parameters
    ----------
    shape : tuple
        the shape of the dataset to be sored / chunked
    dtype : {dtype, string}
        the numpy dtype of the data
    signal_axes: {None, iterable of ints}
        the axes defining "signal space" of the dataset. If None, the default
        h5py chunking is performed.
    """
    typesize = np.dtype(dtype).itemsize
    if signal_axes is None:
        return h5py._hl.filters.guess_chunk(shape, None, typesize)

    # largely based on the guess_chunk in h5py
    CHUNK_MAX = 1024 * 1024
    want_to_keep = multiply([shape[i] for i in signal_axes]) * typesize
    if want_to_keep >= CHUNK_MAX:
        chunks = [1 for _ in shape]
        for i in signal_axes:
            chunks[i] = shape[i]
        return tuple(chunks)

    chunks = [i for i in shape]
    idx = 0
    navigation_axes = tuple(i for i in range(len(shape)) if i not in
                            signal_axes)
    nchange = len(navigation_axes)
    while True:
        chunk_bytes = multiply(chunks) * typesize

        if chunk_bytes < CHUNK_MAX:
            break

        if multiply([chunks[i] for i in navigation_axes]) == 1:
            break
        change = navigation_axes[idx % nchange]
        chunks[change] = np.ceil(chunks[change] / 2.0)
        idx += 1
    return tuple(int(x) for x in chunks)


def overwrite_dataset(group, data, key, signal_axes=None, chunks=None, **kwds):
    if chunks is None:
        if signal_axes is None:
            # Use automatic h5py chunking
            chunks = True
        else:
            # Optimise the chunking to contain at least one signal per chunk
            chunks = get_signal_chunks(data.shape, data.dtype, signal_axes)

    if np.issubdtype(data.dtype, np.dtype('U')):
        # Saving numpy unicode type is not supported in h5py
        data = data.astype(np.dtype('S'))
    if data.dtype == np.dtype('O'):
        # For saving ragged array
        # http://docs.h5py.org/en/stable/special.html#arbitrary-vlen-data
        group.require_dataset(key,
                              chunks,
                              dtype=h5py.special_dtype(vlen=data[0].dtype),
                              **kwds)
        group[key][:] = data[:]

    maxshape = tuple(None for _ in data.shape)

    got_data = False
    while not got_data:
        try:
            these_kwds = kwds.copy()
            these_kwds.update(dict(shape=data.shape,
                                   dtype=data.dtype,
                                   exact=True,
                                   maxshape=maxshape,
                                   chunks=chunks,
                                   shuffle=True,))

            # If chunks is True, the `chunks` attribute of `dset` below
            # contains the chunk shape guessed by h5py
            dset = group.require_dataset(key, **these_kwds)
            got_data = True
        except TypeError:
            # if the shape or dtype/etc do not match,
            # we delete the old one and create new in the next loop run
            del group[key]
    if dset == data:
        # just a reference to already created thing
        pass
    else:
        _logger.info("Chunks used for saving: %s" % str(dset.chunks))
        if isinstance(data, da.Array):
            da.store(data.rechunk(dset.chunks), dset)
        elif data.flags.c_contiguous:
            dset.write_direct(data)
        else:
            dset[:] = data


def hdfgroup2dict(group, dictionary=None, lazy=False):
    if dictionary is None:
        dictionary = {}
    for key, value in group.attrs.items():
        if isinstance(value, bytes):
            value = value.decode()
        if isinstance(value, (np.string_, str)):
            if value == '_None_':
                value = None
        elif isinstance(value, np.bool_):
            value = bool(value)
        elif isinstance(value, np.ndarray) and value.dtype.char == "S":
            # Convert strings to unicode
            value = value.astype("U")
            if value.dtype.str.endswith("U1"):
                value = value.tolist()
        # skip signals - these are handled below.
        if key.startswith('_sig_'):
            pass
        elif key.startswith('_list_empty_'):
            dictionary[key[len('_list_empty_'):]] = []
        elif key.startswith('_tuple_empty_'):
            dictionary[key[len('_tuple_empty_'):]] = ()
        elif key.startswith('_bs_'):
            dictionary[key[len('_bs_'):]] = value.tobytes()
        # The following two elif stataments enable reading date and time from
        # v < 2 of HyperSpy's metadata specifications
        elif key.startswith('_datetime_date'):
            date_iso = datetime.date(
                *ast.literal_eval(value[value.index("("):])).isoformat()
            dictionary[key.replace("_datetime_", "")] = date_iso
        elif key.startswith('_datetime_time'):
            date_iso = datetime.time(
                *ast.literal_eval(value[value.index("("):])).isoformat()
            dictionary[key.replace("_datetime_", "")] = date_iso
        else:
            dictionary[key] = value
    if not isinstance(group, h5py.Dataset):
        for key in group.keys():
            if key.startswith('_sig_'):
                from hyperspy.io import dict2signal
                dictionary[key[len('_sig_'):]] = (
                    dict2signal(hdfgroup2signaldict(
                        group[key], lazy=lazy)))
            elif isinstance(group[key], h5py.Dataset):
                dat = group[key]
                kn = key
                if key.startswith("_list_"):
                    if (h5py.check_string_dtype(dat.dtype) and
                        hasattr(dat, 'asstr')):
                        # h5py 3.0 and newer
                        # https://docs.h5py.org/en/3.0.0/strings.html
                        dat = dat.asstr()[:]
                    ans = np.array(dat)
                    ans = ans.tolist()
                    kn = key[6:]
                elif key.startswith("_tuple_"):
                    ans = np.array(dat)
                    ans = tuple(ans.tolist())
                    kn = key[7:]
                elif dat.dtype.char == "S":
                    ans = np.array(dat)
                    try:
                        ans = ans.astype("U")
                    except UnicodeDecodeError:
                        # There are some strings that must stay in binary,
                        # for example dill pickles. This will obviously also
                        # let "wrong" binary string fail somewhere else...
                        pass
                elif lazy:
                    ans = da.from_array(dat, chunks=dat.chunks)
                else:
                    ans = np.array(dat)
                dictionary[kn] = ans
            elif key.startswith('_hspy_AxesManager_'):
                dictionary[key[len('_hspy_AxesManager_'):]] = AxesManager(
                    [i for k, i in sorted(iter(
                        hdfgroup2dict(
                            group[key], lazy=lazy).items()
                    ))])
            elif key.startswith('_list_'):
                dictionary[key[7 + key[6:].find('_'):]] = \
                    [i for k, i in sorted(iter(
                        hdfgroup2dict(
                            group[key], lazy=lazy).items()
                    ))]
            elif key.startswith('_tuple_'):
                dictionary[key[8 + key[7:].find('_'):]] = tuple(
                    [i for k, i in sorted(iter(
                        hdfgroup2dict(
                            group[key], lazy=lazy).items()
                    ))])
            else:
                dictionary[key] = {}
                hdfgroup2dict(
                    group[key],
                    dictionary[key],
                    lazy=lazy)
    return dictionary


def write_signal(signal, group, **kwds):
    "Writes a hyperspy signal to a hdf5 group"

    group.attrs.update(get_object_package_info(signal))
    if default_version < LooseVersion("1.2"):
        metadata = "mapped_parameters"
        original_metadata = "original_parameters"
    else:
        metadata = "metadata"
        original_metadata = "original_metadata"

    if 'compression' not in kwds:
        kwds['compression'] = 'gzip'

    for axis in signal.axes_manager._axes:
        axis_dict = axis.get_axis_dictionary()
        coord_group = group.create_group(
            'axis-%s' % axis.index_in_array)
        dict2hdfgroup(axis_dict, coord_group, **kwds)
    mapped_par = group.create_group(metadata)
    metadata_dict = signal.metadata.as_dictionary()
    overwrite_dataset(group, signal.data, 'data',
                      signal_axes=signal.axes_manager.signal_indices_in_array,
                      **kwds)
    if default_version < LooseVersion("1.2"):
        metadata_dict["_internal_parameters"] = \
            metadata_dict.pop("_HyperSpy")
    dict2hdfgroup(metadata_dict, mapped_par, **kwds)
    original_par = group.create_group(original_metadata)
    dict2hdfgroup(signal.original_metadata.as_dictionary(), original_par,
                  **kwds)
    learning_results = group.create_group('learning_results')
    dict2hdfgroup(signal.learning_results.__dict__,
                  learning_results, **kwds)
    if hasattr(signal, 'peak_learning_results'):
        peak_learning_results = group.create_group(
            'peak_learning_results')
        dict2hdfgroup(signal.peak_learning_results.__dict__,
                      peak_learning_results, **kwds)

    if len(signal.models):
        model_group = group.file.require_group('Analysis/models')
        dict2hdfgroup(signal.models._models.as_dictionary(),
                      model_group, **kwds)
        for model in model_group.values():
            model.attrs['_signal'] = group.name


def file_writer(filename, signal, *args, **kwds):
    """Writes data to hyperspy's hdf5 format

    Parameters
    ----------
    filename: str
    signal: a BaseSignal instance
    *args, optional
    **kwds, optional
    """
    with h5py.File(filename, mode='w') as f:
        f.attrs['file_format'] = "HyperSpy"
        f.attrs['file_format_version'] = version
        exps = f.create_group('Experiments')
        group_name = signal.metadata.General.title if \
            signal.metadata.General.title else '__unnamed__'
        # / is a invalid character, see #942
        if "/" in group_name:
            group_name = group_name.replace("/", "-")
        expg = exps.create_group(group_name)

        # Add record_by metadata for backward compatibility
        smd = signal.metadata.Signal
        if signal.axes_manager.signal_dimension == 1:
            smd.record_by = "spectrum"
        elif signal.axes_manager.signal_dimension == 2:
            smd.record_by = "image"
        else:
            smd.record_by = ""
        try:
            write_signal(signal, expg, **kwds)
        except BaseException:
            raise
        finally:
            del smd.record_by



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import os
from datetime import datetime as dt
import warnings
import logging
from distutils.version import LooseVersion


# Plugin characteristics
# ----------------------
format_name = 'Protochips'
description = 'Reads Protochips log files (heating/baising and gas cell)'
full_support = False
# Recognised file extension
file_extensions = ['csv', 'CSV']
default_extension = 0
# Reading capabilities
reads_images = False
reads_spectrum = False
reads_spectrum_image = False
# Writing capabilities
writes = False

_logger = logging.getLogger(__name__)


# At some point, if there is another readerw, whith also use csv file, it will
# be necessary to mention the other reader in this message (and to add an
# argument in the load function to specify the correct reader)
invalid_file_error = "The Protochips csv reader can't import the file, please"\
    " make sure, this is a valid Protochips log file."


def file_reader(filename, *args, **kwds):
    csv_file = ProtochipsCSV(filename)
    return _protochips_log_reader(csv_file)


def _protochips_log_reader(csv_file):
    csvs = []
    for key in csv_file.logged_quantity_name_list:
        try:
            csvs.append(csv_file.get_dictionary(key))
        except BaseException:
            raise IOError(invalid_file_error)
    return csvs


class ProtochipsCSV(object):

    def __init__(self, filename, ):
        self.filename = filename
        self._parse_header()
        self._read_data()

    def _parse_header(self):
        with open(self.filename, 'r') as f:
            s = f.readline()
            self.column_name = s.replace(', ', ',').replace('\n', '').split(',')
            if not self._is_protochips_csv_file():
                raise IOError(invalid_file_error)
            self._read_all_metadata_header(f)
        self.logged_quantity_name_list = self.column_name[2:]

    def _is_protochips_csv_file(self):
        # This check is not great, but it's better than nothing...
        if 'Time' in self.column_name and 'Notes' in self.column_name and len(
                self.column_name) >= 3:
            return True
        else:
            return False

    def get_dictionary(self, quantity):
        return {'data': self._data_dictionary[quantity],
                'axes': self._get_axes(),
                'metadata': self._get_metadata(quantity),
                'mapping': self._get_mapping(),
                'original_metadata': {'Protochips_header':
                                      self._get_original_metadata()}}

    def _get_original_metadata(self):
        d = {'Start time': self.start_datetime}
        d['Time units'] = self.time_units
        for quantity in self.logged_quantity_name_list:
            d['%s_units' % quantity] = self._parse_quantity_units(quantity)
        if self.user:
            d['User'] = self.user
        d['Calibration file path'] = self._parse_calibration_filepath()
        d['Time axis'] = self._get_metadata_time_axis()
        # Add the notes here, because there are not well formatted enough to
        # go in metadata
        d['Original notes'] = self._parse_notes()
        return d

    def _get_metadata(self, quantity):
        date, time = np.datetime_as_string(self.start_datetime).split('T')
        return {'General': {'original_filename': os.path.split(self.filename)[1],
                            'title': '%s (%s)' % (quantity,
                                                  self._parse_quantity_units(quantity)),
                            'date': date,
                            'time': time},
                "Signal": {'signal_type': '',
                           'quantity': self._parse_quantity(quantity)}}

    def _get_mapping(self):
        mapping = {
            "Protochips_header.Calibration file path": (
                "General.notes",
                self._parse_calibration_file_name),
            "Protochips_header.User": (
                "General.authors",
                None),
        }
        return mapping

    def _get_metadata_time_axis(self):
        return {'value': self.time_axis,
                'units': self.time_units}

    def _read_data(self):
        names = [name.replace(' ', '_') for name in self.column_name]
        # Necessary for numpy >= 1.14
        kwargs = {'encoding': 'latin1'} if np.__version__ >= LooseVersion("1.14") else {
        }
        data = np.genfromtxt(self.filename, delimiter=',', dtype=None,
                             names=names,
                             skip_header=self.header_last_line_number,
                             unpack=True, **kwargs)

        self._data_dictionary = dict()
        for i, name, name_dtype in zip(range(len(names)), self.column_name,
                                       names):
            if name == 'Notes':
                self.notes = data[name_dtype].astype(str)
            elif name == 'Time':
                self.time_axis = data[name_dtype]
            else:
                self._data_dictionary[name] = data[name_dtype]

    def _parse_notes(self):
        arr = np.vstack((self.time_axis, self.notes))
        return np.compress(arr[1] != '', arr, axis=1)

    def _parse_calibration_filepath(self):
         # for the gas cell, the calibration is saved in the notes colunm
        if hasattr(self, "calibration_file"):
            calibration_file = self.calibration_file
        else:
            calibration_file = "The calibration files names are saved in the"\
                " 'Original notes' array of the original metadata."
        return calibration_file

    def _parse_calibration_file_name(self, path):
        basename = os.path.basename(path)
        return "Calibration file name: %s" % basename.split('\\')[-1]

    def _get_axes(self):
        scale = np.diff(self.time_axis[1:-1]).mean()
        max_diff = np.diff(self.time_axis[1:-1]).max()
        units = 's'
        offset = 0
        if self.time_units == 'Milliseconds':
            scale /= 1000
            max_diff /= 1000
            # Once we support non-linear axis, don't forgot to update the
            # documentation of the protochips reader
            _logger.warning("The time axis is not linear, the time step is "
                            "thus extrapolated to {0} {1}. The maximal step in time step is {2} {1}".format(
                                scale, units, max_diff))
        else:
            warnings.warn("Time units not recognised, assuming second.")

        return [{'size': self.time_axis.shape[0],
                 'index_in_array': 0,
                 'name': 'Time',
                 'scale': scale,
                 'offset': offset,
                 'units': units,
                 'navigate': False
                 }]

    def _parse_quantity(self, quantity):
        quantity_name = quantity.split(' ')[-1]
        return '%s (%s)' % (quantity_name,
                            self._parse_quantity_units(quantity))

    def _parse_quantity_units(self, quantity):
        quantity = quantity.split(' ')[-1].lower()
        return self.__dict__['%s_units' % quantity]

    def _read_all_metadata_header(self, f):
        param, value = self._parse_metadata_header(f.readline())
        i = 2
        while 'User' not in param:  # user should be the last of the header
            if 'Calibration file' in param:
                self.calibration_file = value
            elif 'Date (yyyy.mm.dd)' in param:
                date = value
            elif 'Time (hh:mm:ss.ms)' in param:
                time = value
            else:
                attr_name = param.replace(' ', '_').lower()
                self.__dict__[attr_name] = value
            i += 1
            try:
                param, value = self._parse_metadata_header(f.readline())
            except ValueError:
                # when the last line of header does not contain 'User',
                # possibly some old file.
                self.user = None
                break
            except IndexError:
                _logger.warning("The metadata may not be parsed properly.")
                break
        else:
            self.user = value
        self.header_last_line_number = i
        self.start_datetime = np.datetime64(dt.strptime(date + time,
                                                        "%Y.%m.%d%H:%M:%S.%f"))

    def _parse_metadata_header(self, line):
        return line.replace(', ', ',').split(',')[1].split(' = ')



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

"""
The Signal class and its specialized subclasses:

    BaseSignal
        For generic data with arbitrary signal_dimension. All other signal
        classes inherit from this one. It should only be used with none of
        the others is appropriated.
    Signal1D
        For generic data with signal_dimension equal 1, i.e. spectral data of
        n-dimensions. The signal is unbinned by default.
    Signal2D
        For generic data with signal_dimension equal 2, i.e. image data of
        n-dimensions. The signal is unbinned by default.
    ComplexSignal
        For generic complex data with arbitrary signal_dimension.
    ComplexSignal1D
        For generic complex data with signal_dimension equal 1, i.e. spectral
        data of n-dimensions. The signal is unbinned by default.
    ComplexSignal2D
        For generic complex data with signal_dimension equal 2, i.e. image
        data of n-dimensions. The signal is unbinned by default.
    EELSSpectrum
        For electron energy-loss data with signal_dimension equal 1, i.e.
        spectral data of n-dimensions. The signal is binned by default.
    EDSTEMSpectrum
        For electron energy-dispersive X-rays data acquired in a transmission
        electron microscopy with signal_dimension equal 1, i.e.
        spectral data of n-dimensions. The signal is binned by default.
    EDSSEMSpectrum
        For electron energy-dispersive X-rays data acquired in a scanning
        electron microscopy with signal_dimension equal 1, i.e.
        spectral data of n-dimensions. The signal is binned by default.
    DielectricFunction
        For dielectric function data with signal_dimension equal 1. The signal
        is unbinned by default.
    HolographyImage
        For 2D-images taken via electron holography. Electron wave as
        ComplexSignal2D can be reconstructed from them.
"""

# -*- coding: utf-8 -*-
from hyperspy.extensions import EXTENSIONS as _EXTENSIONS
import importlib

_g = globals()
for _signal, _specs in _EXTENSIONS["signals"].items():
    if not _specs["lazy"]:
        _g[_signal] = getattr(
            importlib.import_module(
                _specs["module"]), _signal)

del importlib



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <http://www.gnu.org/licenses/>.

import pytest

from hyperspy.datasets.example_signals import EDS_SEM_Spectrum
from hyperspy.misc.model_tools import (_is_iter, _iter_join, _non_iter,
                                       current_component_values,
                                       current_model_values)


class TestSetParameters:

    def setup_method(self):
        self.model = EDS_SEM_Spectrum().create_model()
        self.component = self.model[1]
        # We use bmin instead of A because it's a bit more exotic
        self.component.A.bmin = 1.23456789012
        self.component_not_free = self.model[3]
        self.component_not_free.set_parameters_not_free()
        self.component_not_free.A.bmin = 9.87654321098
        self.component_inactive = self.model[4]
        self.component_inactive.active = False
        self.component_inactive.A.bmin = 5.67890123456

    @pytest.mark.parametrize("only_free, only_active", [(True, False), (True, False)])
    def test_component_current_component_values(self, only_free, only_active):
        "Many decimals aren't printed, few decimals are"
        string_representation = str(current_component_values(self.component, only_free, only_active).__repr__())
        html_representation = str(current_component_values(self.component, only_free, only_active)._repr_html_())
        assert "1.234" in string_representation
        assert "1.23456789012" not in string_representation
        assert "1.234" in html_representation
        assert "1.23456789012" not in html_representation

    def test_component_current_component_values_only_free(self, only_free=True, only_active=False):
        "Parameters with free=False values should not be present in repr"
        string_representation = str(current_component_values(self.component_not_free, only_free, only_active).__repr__())
        html_representation = str(current_component_values(self.component_not_free, only_free, only_active)._repr_html_())
        assert "9.87" not in string_representation
        assert "9.87" not in html_representation

    @pytest.mark.parametrize("only_free, only_active", [(True, False), (True, False)])
    def test_component_current_model_values(self, only_free, only_active):
        "Many decimals aren't printed, few decimals are"
        string_representation = str(current_model_values(self.model, only_free, only_active).__repr__())
        html_representation = str(current_model_values(self.model, only_free, only_active)._repr_html_())
        assert "1.234" in string_representation
        assert "1.23456789012" not in string_representation
        assert "1.234" in html_representation
        assert "1.23456789012" not in html_representation

        if only_free:
            # Parameters with free=False values should not be present in repr
            assert "9.87" not in string_representation
            assert "9.87" not in html_representation
        if only_active:
            # components with active=False values should not be present in repr"
            assert "5.67" not in string_representation
            assert "5.67" not in html_representation

    @pytest.mark.parametrize("only_free, only_active", [(True, False), (True, False)])
    def test_component_current_model_values_comp_list(self, only_free, only_active):
        comp_list = [self.component, self.component_not_free, self.component_inactive]
        string_representation = str(current_model_values(self.model, only_free, only_active, comp_list).__repr__())
        html_representation = str(current_model_values(self.model, only_free, only_active, comp_list)._repr_html_())
        assert "1.234" in string_representation
        assert "1.23456789012" not in string_representation
        assert "1.234" in html_representation
        assert "1.23456789012" not in html_representation

        if only_free:
            assert "9.87" not in string_representation
            assert "9.87" not in html_representation
        if only_active:
            assert "5.67" not in string_representation
            assert "5.67" not in html_representation

    @pytest.mark.parametrize("fancy", (True, False))
    def test_model_current_model_values(self, fancy):
        self.model.print_current_values(fancy=fancy)
        
    @pytest.mark.parametrize("fancy", (True, False))
    def test_component_print_current_values(self, fancy):
        self.model[0].print_current_values(fancy=fancy)

    @pytest.mark.parametrize("fancy", (True, False))
    def test_model_print_current_values(self, fancy):
        self.model.print_current_values(fancy=fancy)

    def test_zero_in_fancy_print(self):
        "Ensure parameters with value=0 are printed too"
        assert "<td>a1</td><td>True</td><td>     0</td>" in current_component_values(self.model[0])._repr_html_()

    def test_zero_in_normal_print(self):
        "Ensure parameters with value=0 are printed too"
        assert "            a0 |  True |          0 |" in str(current_component_values(self.model[0]).__repr__)

    def test_related_tools(self):
        assert _is_iter([1,2,3])
        assert _is_iter((1,2,3))
        assert not _is_iter(1)

        assert _iter_join([1.2345678, 5.67890]) == '(1.23457, 5.6789)'
        assert _iter_join([1.2345678, 5.67890]) == '(1.23457, 5.6789)'
        assert _iter_join([1, 5]) == '(     1,      5)'

        assert _non_iter(None) == ""
        assert _non_iter(5) == '     5'
        assert _non_iter(5.123456789) == '5.12346'



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

"""Common docstring snippets for parameters.

"""

FUNCTION_ND_DOCSTRING = \
    """Returns a numpy array containing the value of the component for all
        indices. If enough memory is available, this is useful to quickly to
        obtain the fitted component without iterating over the navigation axes.
        """



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import os
import struct

import numpy as np
import pytest
import traits.api as t

import hyperspy.api as hs

DATA_DIR = os.path.join(os.path.dirname(__file__), "empad_data")
FILENAME_STACK_RAW = os.path.join(DATA_DIR, "series_x10.raw")
FILENAME_MAP_RAW = os.path.join(DATA_DIR, "scan_x4_y4.raw")


def _create_raw_data(filename, shape):
    size = np.prod(shape)
    data = np.arange(size).reshape(shape).astype("float32")
    data.tofile(filename)


def setup_module():
    _create_raw_data(FILENAME_STACK_RAW, (166400,))
    _create_raw_data(FILENAME_MAP_RAW, (4 * 4 * 130 * 128))


def teardown_module():
    fs = [f for f in [FILENAME_STACK_RAW, FILENAME_MAP_RAW] if os.path.exists(f)]

    for f in fs:
        try:
            os.remove(f)
        except Exception as e:
            print(e)


@pytest.mark.parametrize("lazy", (False, True))
def test_read_stack(lazy):
    s = hs.load(os.path.join(DATA_DIR, "stack_images.xml"), lazy=lazy)
    assert s.data.dtype == "float32"
    ref_data = np.arange(166400).reshape((10, 130, 128))[..., :128, :]
    np.testing.assert_allclose(s.data, ref_data.astype("float32"))
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "width"
    assert signal_axes[1].name == "height"
    for axis in signal_axes:
        assert axis.units == t.Undefined
        assert axis.scale == 1.0
        assert axis.offset == -64
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "series_count"
    assert navigation_axes[0].units == "ms"
    assert navigation_axes[0].scale == 1.0
    assert navigation_axes[0].offset == 0.0

    assert s.metadata.General.date == "2019-06-07"
    assert s.metadata.General.time == "13:17:22.590279"
    assert s.metadata.Signal.signal_type == "electron_diffraction"


@pytest.mark.parametrize("lazy", (False, True))
def test_read_map(lazy):
    s = hs.load(os.path.join(DATA_DIR, "map4x4.xml"), lazy=lazy)
    assert s.data.dtype == "float32"
    ref_data = np.arange(266240).reshape((4, 4, 130, 128))[..., :128, :]
    np.testing.assert_allclose(s.data, ref_data.astype("float32"))
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "width"
    assert signal_axes[1].name == "height"
    for axis in signal_axes:
        assert axis.units == "1/nm"
        np.testing.assert_allclose(axis.scale, 0.1826537)
        np.testing.assert_allclose(axis.offset, -11.689837)
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "scan_y"
    assert navigation_axes[1].name == "scan_x"
    for axis in navigation_axes:
        assert axis.units == "µm"
        np.testing.assert_allclose(axis.scale, 1.1415856)
        np.testing.assert_allclose(axis.offset, 0.0)

    assert s.metadata.General.date == "2019-06-06"
    assert s.metadata.General.time == "13:30:00.164675"
    assert s.metadata.Signal.signal_type == "electron_diffraction"



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from scipy.stats import norm

from hyperspy.signals import Signal2D, BaseSignal
from hyperspy._signals.lazy import LazySignal
from hyperspy.decorators import lazifyTestClass
from hyperspy.signal_tools import PeaksFinder2D


def _generate_dataset():
    coefficients = np.array(
        [350949.04890400 + 0.j, -22003.98742841 + 51494.56650429j,
         37292.52741553 + 38067.97686711j, 37292.52741553 - 38067.97686711j,
         -22003.98742841 - 51494.56650429j]
    )
    coordinates = np.array([[0, 26, 30, 994, 998],
                            [0, 1003, 39, 985, 21]]
    )
    dense = np.zeros((1024, 1024), dtype=complex)
    dense[coordinates[0], coordinates[1]] = coefficients
    dense = Signal2D(np.real(np.fft.ifft2(dense)))
    dense = dense.isig[500:550, 500:550]

    coefficients = np.array(
        [10, 5, 86, 221, 6, 95, 70, 12, 255, 5, 255, 3, 23,
         24, 77, 255, 11, 255, 8, 35, 195, 165, 27, 255, 8, 14,
         255, 21, 53, 107, 255, 18, 255, 4, 26, 255, 39, 27, 255,
         6, 255, 7, 13, 37, 35, 9, 83]
    )
    coordinates = np.array(
        [[3, 40],    [3, 138],  [9, 67],   [14, 95],   [20, 23],
         [20, 122],  [26, 51],  [26, 100], [31, 78],   [31, 128],
         [37, 107],  [38, 7],   [43, 34],  [43, 84],   [43, 134],
         [49, 62],   [49, 112], [54, 90],  [60, 17],   [60, 67],
         [60, 118],  [66, 45],  [66, 96],  [72, 73],   [72, 124],
         [77, 51],   [77, 101], [83, 28],  [83, 79],   [83, 130],
         [89, 57],   [89, 107], [95, 85],  [101, 12],  [101, 62],
         [101, 113], [106, 40], [107, 91], [112, 68],  [113, 119],
         [119, 97],  [124, 23], [124, 74], [124, 125], [130, 51],
         [130, 103], [136, 80]])
    sparse = np.zeros((144, 144))
    xs, ys = np.ogrid[:144, :144]
    for (x0, y0), a in zip(coordinates, coefficients):
        sparse += a * norm.pdf(xs, x0)*norm.pdf(ys, y0)
    sparse = sparse[50:100, 50:100]
    sparse_nav0d = Signal2D(sparse)
    sparse_nav1d = Signal2D(np.stack([sparse]*2))
    sparse_nav2d = Signal2D(np.stack([[sparse]*2]*3))
    shifts = np.array([[2*i, 2*i] for i in range(sparse_nav2d.axes_manager.navigation_size)])
    sparse_nav2d_shifted = sparse_nav2d.deepcopy()
    sparse_nav2d_shifted.align2D(shifts=shifts, fill_value=0)

    return dense, sparse_nav0d, sparse_nav1d, sparse_nav2d, sparse_nav2d_shifted


def _generate_reference():
    xref, yref = 72, 72
    ref = np.zeros((144, 144))
    xs, ys = np.ogrid[:144, :144]
    ref += 100 * norm.pdf(xs, xref)*norm.pdf(ys, yref)
    return Signal2D(ref), xref, yref


PEAK_METHODS = ['local_max', 'max', 'minmax', 'zaefferer', 'stat',
                'laplacian_of_gaussian', 'difference_of_gaussian',
                'template_matching']
DATASETS = _generate_dataset()
DATASETS_NAME = ["dense", "sparse_nav0d", "sparse_nav1d", "sparse_nav2d"]


@lazifyTestClass
class TestFindPeaks2D:

    def setup_method(self, method):
        # All these signal needs to be in the `setup_method` to get "lazified"
        self.dense = DATASETS[0]
        self.sparse_nav0d = DATASETS[1]
        self.sparse_nav1d = DATASETS[2]
        self.sparse_nav2d = DATASETS[3]
        self.sparse_nav2d_shifted = DATASETS[4]
        self.ref, self.xref, self.yref = _generate_reference()

    @pytest.mark.parametrize('method', PEAK_METHODS)
    @pytest.mark.parametrize('dataset_name', DATASETS_NAME)
    @pytest.mark.parametrize('parallel', [True, False])
    def test_find_peaks(self, method, dataset_name, parallel):
        if method == 'stat':
            pytest.importorskip("sklearn")
        dataset = getattr(self, dataset_name)
        # Parallel is not used in `map` for lazy signal
        if parallel and dataset._lazy:
            pytest.skip("Parallel=True is ignored for lazy signal.")

        if method == 'template_matching':
            disc = np.zeros((11, 11))
            disc[2:9, 2:9] = 0.5
            disc[4:7, 4:7] = 0.75
            disc[5, 5] = 1
            peaks = dataset.find_peaks(method=method, parallel=parallel,
                                       interactive=False, template=disc)
        else:
            peaks = dataset.find_peaks(method=method, parallel=parallel,
                                       interactive=False)
        assert isinstance(peaks, BaseSignal)
        assert not isinstance(peaks, LazySignal)

        # Check navigation shape
        np.testing.assert_equal(dataset.axes_manager.navigation_shape,
                        peaks.axes_manager.navigation_shape)
        if dataset.axes_manager.navigation_size == 0:
            shape = (1,)
        else:
            shape = dataset.axes_manager.navigation_shape[::-1]
        assert peaks.data.shape == shape
        assert peaks.data[0].shape[-1] == 2

    @pytest.mark.parametrize('parallel', [True, False])
    def test_ordering_results(self, parallel):
        peaks = self.sparse_nav2d_shifted.find_peaks(parallel=parallel,
                                                     interactive=False)

        np.testing.assert_equal(peaks.inav[0, 0].data,
                        np.array([[27,  1],
                                  [10, 17],
                                  [22, 23],
                                  [33, 29]]))
        np.testing.assert_equal(peaks.inav[0, 1].data,
                        np.array([[35,  3],
                                  [ 6, 13],
                                  [18, 19],
                                  [29, 25]]))

    @pytest.mark.parametrize('method', PEAK_METHODS)
    @pytest.mark.parametrize('parallel', [True, False])
    def test_gets_right_answer(self, method, parallel):
        if method == 'stat':
            pytest.importorskip("sklearn")
        ans = np.empty((1,), dtype=object)
        ans[0] = np.array([[self.xref, self.yref]])
        if method == 'template_matching':
            disc = np.zeros((5, 5))
            disc[1:4, 1:4] = 0.5
            disc[2,2] = 1
            peaks = self.ref.find_peaks(method=method, parallel=parallel,
                                        interactive=False, template=disc)
        else:
            peaks = self.ref.find_peaks(method=method, parallel=parallel,
                                        interactive=False)
        np.testing.assert_allclose(peaks.data[0], ans[0])

    def test_return_peaks(self):
        sig = self.sparse_nav2d_shifted
        axes_dict = sig.axes_manager._get_axes_dicts(
            sig.axes_manager.navigation_axes)
        peaks = BaseSignal(np.empty(sig.axes_manager.navigation_shape),
                           axes=axes_dict)
        pf2D = PeaksFinder2D(sig, method='local_max', peaks=peaks)
        np.testing.assert_allclose(peaks.data, np.array([[22, 23]]))

        pf2D.local_max_threshold = 2
        pf2D._update_peak_finding()
        result_index0 = np.array([[10, 17], [22, 23], [33, 29]])
        np.testing.assert_allclose(peaks.data, result_index0)
        pf2D.compute_navigation()
        pf2D.close()

        assert peaks.data.shape == (3, 2)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np

from hyperspy.io import load

MY_PATH = os.path.dirname(__file__)

header_keys = ['H01_Signature',
                 'H02_Format',
                 'H03_Number_of_Objects',
                 'H04_Version',
                 'H05_Object_Type',
                 'H06_Object_Name',
                 'H07_Operator_Name',
                 'H08_P_Size',
                 'H09_Acquisition_Type',
                 'H10_Range_Type',
                 'H11_Special_Points',
                 'H12_Absolute',
                 'H13_Gauge_Resolution',
                 'H14_W_Size',
                 'H15_Size_of_Points',
                 'H16_Zmin',
                 'H17_Zmax',
                 'H18_Number_of_Points',
                 'H19_Number_of_Lines',
                 'H20_Total_Nb_of_Pts',
                 'H21_X_Spacing',
                 'H22_Y_Spacing',
                 'H23_Z_Spacing',
                 'H24_Name_of_X_Axis',
                 'H25_Name_of_Y_Axis',
                 'H26_Name_of_Z_Axis',
                 'H27_X_Step_Unit',
                 'H28_Y_Step_Unit',
                 'H29_Z_Step_Unit',
                 'H30_X_Length_Unit',
                 'H31_Y_Length_Unit',
                 'H32_Z_Length_Unit',
                 'H33_X_Unit_Ratio',
                 'H34_Y_Unit_Ratio',
                 'H35_Z_Unit_Ratio',
                 'H36_Imprint',
                 'H37_Inverted',
                 'H38_Levelled',
                 'H39_Obsolete',
                 'H40_Seconds',
                 'H41_Minutes',
                 'H42_Hours',
                 'H43_Day',
                 'H44_Month',
                 'H45_Year',
                 'H46_Day_of_week',
                 'H47_Measurement_duration',
                 'H48_Compressed_data_size',
                 'H49_Obsolete',
                 'H50_Comment_size',
                 'H51_Private_size',
                 'H52_Client_zone',
                 'H53_X_Offset',
                 'H54_Y_Offset',
                 'H55_Z_Offset',
                 'H56_T_Spacing',
                 'H57_T_Offset',
                 'H58_T_Axis_Name',
                 'H59_T_Step_Unit',
                 'H60_Comment',
                 ]

atto_head_keys = ['WAFER',
                            'SITE IMAGE',
                            'SEM',
                            'CHANNELS',
                            'SPECTROMETER',
                            'SCAN',
                            ]

atto_wafer_keys = ['Lot Number',
                     'ID',
                     'Type',
                     'Center Position X',
                     'Center Position X_units',
                     'Center Position Y',
                     'Center Position Y_units',
                     'Orientation',
                     'Orientation_units',
                     'Diameter',
                     'Diameter_units',
                     'Flat Length',
                     'Flat Length_units',
                     'Edge Exclusion',
                     'Edge Exclusion_units',
                     ]

atto_scan_keys = ['Mode',
                 'HYP Dwelltime',
                 'HYP Dwelltime_units',
                 'Resolution_X',
                 'Resolution_X_units',
                 'Resolution_Y',
                 'Resolution_Y_units',
                 'Reference_Size_X',
                 'Reference_Size_Y',
                 'Voltage Calibration Range_X',
                 'Voltage Calibration Range_X_units',
                 'Voltage Calibration Range_Y',
                 'Voltage Calibration Range_Y_units',
                 'Start_X',
                 'Size_X',
                 'Start_Y',
                 'Size_Y',
                 'Rotate',
                 'Rotate_units',
                 ]

def test_load_profile():
    #Signal loading
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_profile.pro")
    s = load(fname)

    #Verifying signal shape and axes dimensions, navigation (not data themselves)
    assert s.data.shape == (128,)
    assert s.data.dtype == np.dtype(float)
    np.testing.assert_allclose(s.axes_manager[0].scale,8.252197e-05)
    np.testing.assert_allclose(s.axes_manager[0].offset,0.0)
    assert s.axes_manager[0].name == 'Width'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[0].size == 128
    assert s.axes_manager[0].navigate == False

    #Metadata verification
    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'

    #Original metadata. We verify that the correct structure is given
    #and the right headers but not the values
    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0']
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                            == header_keys

def test_load_RGB():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_RGB.sur")
    s = load(fname)
    assert s.data.shape == (200, 200)
    assert s.data.dtype == np.dtype([('R', 'u1'), ('G', 'u1'), ('B', 'u1')])

    np.testing.assert_allclose(s.axes_manager[0].scale,0.35277777)
    np.testing.assert_allclose(s.axes_manager[0].offset,208.8444519)
    np.testing.assert_allclose(s.axes_manager[1].scale,0.35277777)
    np.testing.assert_allclose(s.axes_manager[1].offset,210.608337)
    assert s.axes_manager[0].name == 'X'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[1].name == 'Y'
    assert s.axes_manager[1].units == 'mm'
    assert s.axes_manager[0].size == 200
    assert s.axes_manager[0].navigate == False
    assert s.axes_manager[1].size == 200
    assert s.axes_manager[1].navigate == False

    md = s.metadata
    assert md.Signal.quantity == 'Z'

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0',
                                                'Object_0_Channel_1',
                                                'Object_0_Channel_2']
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) == header_keys

def test_load_spectra():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_spectra.pro")
    s = load(fname)

    assert s.data.shape == (65, 512)
    assert s.data.dtype == np.dtype('float64')

    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    np.testing.assert_allclose(s.axes_manager[0].scale,0.00011458775406936184)
    np.testing.assert_allclose(s.axes_manager[0].offset,0.0)
    np.testing.assert_allclose(s.axes_manager[1].scale,1.084000246009964e-06)
    np.testing.assert_allclose(s.axes_manager[1].offset,0.00017284281784668565)
    assert s.axes_manager[0].name == 'Spectrum positi'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[1].name == 'Wavelength'
    assert s.axes_manager[1].units == 'mm'
    assert s.axes_manager[0].size == 65
    assert s.axes_manager[0].navigate == True
    assert s.axes_manager[1].size == 512
    assert s.axes_manager[1].navigate == False

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0',]
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                                == header_keys

def test_load_spectral_map_compressed():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_spectral_map_compressed.sur")
    s = load(fname)

    assert s.data.shape == (12, 10, 281)
    assert s.data.dtype == np.dtype('float64')

    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    np.testing.assert_allclose(s.axes_manager[0].scale,8.252198e-05)
    np.testing.assert_allclose(s.axes_manager[0].offset,0.005694016348570585)
    np.testing.assert_allclose(s.axes_manager[1].scale,8.252198e-05)
    np.testing.assert_allclose(s.axes_manager[1].offset,0.0054464503191411495)
    np.testing.assert_allclose(s.axes_manager[2].scale,1.084000246009964e-06)
    np.testing.assert_allclose(s.axes_manager[2].offset,0.00034411484375596046)
    assert s.axes_manager[0].name == 'Width'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[1].name == 'Height'
    assert s.axes_manager[1].units == 'mm'
    assert s.axes_manager[2].name == 'Wavelength'
    assert s.axes_manager[2].units == 'mm'
    assert s.axes_manager[0].size == 10
    assert s.axes_manager[0].navigate == True
    assert s.axes_manager[1].size == 12
    assert s.axes_manager[1].navigate == True
    assert s.axes_manager[2].size == 281
    assert s.axes_manager[2].navigate == False

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0',]
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) \
                                                            == ['Header','Parsed']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                            == header_keys

    assert list(omd.Object_0_Channel_0.Parsed.as_dictionary().keys()) \
                                                            == atto_head_keys

    assert list(omd.Object_0_Channel_0.Parsed.WAFER.as_dictionary().keys()) \
                                                            == atto_wafer_keys

    assert list(omd.Object_0_Channel_0.Parsed.SCAN.as_dictionary().keys()) \
                                                            == atto_scan_keys

def test_load_spectral_map():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_spectral_map.sur")
    s = load(fname)

    assert s.data.shape == (12, 10, 310)
    assert s.data.dtype == np.dtype('float64')

    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    np.testing.assert_allclose(s.axes_manager[0].scale,8.252197585534304e-05)
    np.testing.assert_allclose(s.axes_manager[0].offset,0.00701436772942543)
    np.testing.assert_allclose(s.axes_manager[1].scale,8.252197585534304e-05)
    np.testing.assert_allclose(s.axes_manager[1].offset,0.003053313121199608)
    np.testing.assert_allclose(s.axes_manager[2].scale,1.084000246009964e-6)
    np.testing.assert_allclose(s.axes_manager[2].offset,0.0003332748601678759)
    assert s.axes_manager[0].name == 'Width'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[1].name == 'Height'
    assert s.axes_manager[1].units == 'mm'
    assert s.axes_manager[2].name == 'Wavelength'
    assert s.axes_manager[2].units == 'mm'
    assert s.axes_manager[0].size == 10
    assert s.axes_manager[0].navigate == True
    assert s.axes_manager[1].size == 12
    assert s.axes_manager[1].navigate == True
    assert s.axes_manager[2].size == 310
    assert s.axes_manager[2].navigate == False


    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0',]
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header','Parsed']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                               == header_keys

    assert list(omd.Object_0_Channel_0.Parsed.as_dictionary().keys()) \
                                                            == atto_head_keys

    assert list(omd.Object_0_Channel_0.Parsed.WAFER.as_dictionary().keys()) \
                                                            == atto_wafer_keys

    assert list(omd.Object_0_Channel_0.Parsed.SCAN.as_dictionary().keys())  \
                                                            == atto_scan_keys

def test_load_spectrum_compressed():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_spectrum_compressed.pro")
    s = load(fname)
    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    assert s.data.shape == (512,)
    #np.testing.assert_allclose(s.axes_manager[0].scale,1.0)
    #np.testing.assert_allclose(s.axes_manager[0].offset,0.0)
    np.testing.assert_allclose(s.axes_manager[0].scale,1.084000246009964e-6)
    np.testing.assert_allclose(s.axes_manager[0].offset,172.84281784668565e-6)

    #assert s.axes_manager[0].name == 'T'
    #assert s.axes_manager[0].units == ''
    assert s.axes_manager[0].name == 'Wavelength'
    assert s.axes_manager[0].units == 'mm'
    #assert s.axes_manager[0].size == 1
    #assert s.axes_manager[0].navigate == True
    assert s.axes_manager[0].size == 512
    assert s.axes_manager[0].navigate == False

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0']
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                                == header_keys

def test_load_spectrum():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_spectrum.pro")
    s = load(fname)
    assert s.data.shape == (512,)

    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    #np.testing.assert_allclose(s.axes_manager[0].scale,1.0)
    #np.testing.assert_allclose(s.axes_manager[0].offset,0.0)
    np.testing.assert_allclose(s.axes_manager[0].scale,1.084000246009964e-6)
    np.testing.assert_allclose(s.axes_manager[0].offset,172.84281784668565e-6)

    #assert s.axes_manager[0].name == 'T'
    #assert s.axes_manager[0].units == ''
    assert s.axes_manager[0].name == 'Wavelength'
    assert s.axes_manager[0].units == 'mm'
    #assert s.axes_manager[0].size == 1
    #assert s.axes_manager[0].navigate == True
    assert s.axes_manager[0].size == 512
    assert s.axes_manager[0].navigate == False

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0']
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                                == header_keys

def test_load_surface():
    fname = os.path.join(MY_PATH, "sur_data",
                         "test_surface.sur")
    s = load(fname)
    md = s.metadata
    assert md.Signal.quantity == 'CL Intensity (a.u.)'
    assert s.data.shape == (128,128)
    np.testing.assert_allclose(s.axes_manager[0].scale,8.252198e-05)
    np.testing.assert_allclose(s.axes_manager[0].offset,0.0)
    np.testing.assert_allclose(s.axes_manager[1].scale,8.252198e-05)
    np.testing.assert_allclose(s.axes_manager[1].offset,0.0)

    assert s.axes_manager[0].name == 'Width'
    assert s.axes_manager[0].units == 'mm'
    assert s.axes_manager[1].name == 'Height'
    assert s.axes_manager[1].units == 'mm'
    assert s.axes_manager[0].size == 128
    assert s.axes_manager[0].navigate == False
    assert s.axes_manager[1].size == 128
    assert s.axes_manager[1].navigate == False

    omd = s.original_metadata
    assert list(omd.as_dictionary().keys()) == ['Object_0_Channel_0']
    assert list(omd.Object_0_Channel_0.as_dictionary().keys()) == ['Header']
    assert list(omd.Object_0_Channel_0.Header.as_dictionary().keys()) \
                                                            == header_keys



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from dateutil import parser, tz

import hyperspy.misc.date_time_tools as dtt
from hyperspy.misc.test_utils import assert_deep_almost_equal
from hyperspy.misc.utils import DictionaryTreeBrowser


def _get_example(date, time, time_zone=None):
    md = DictionaryTreeBrowser({'General': {'date': date,
                                            'time': time}})
    if time_zone:
        md.set_item('General.time_zone', time_zone)
        dt = parser.parse('%sT%s' % (date, time))
        dt = dt.replace(tzinfo=tz.gettz(time_zone))
        iso = dt.isoformat()
    else:
        iso = '%sT%s' % (date, time)
        dt = parser.parse(iso)
    return md, dt, iso


md1, dt1, iso1 = _get_example('2014-12-27', '00:00:00', 'UTC')
serial1 = 42000.00

md2, dt2, iso2 = _get_example('2124-03-25', '10:04:48', 'EST')
serial2 = 81900.62833333334

md3, dt3, iso3 = _get_example('2016-07-12', '22:57:32')
serial3 = 42563.95662037037


def test_get_date_time_from_metadata():
    assert (dtt.get_date_time_from_metadata(md1) ==
            '2014-12-27T00:00:00+00:00')
    assert (dtt.get_date_time_from_metadata(md1, formatting='ISO') ==
            '2014-12-27T00:00:00+00:00')
    assert (dtt.get_date_time_from_metadata(md1, formatting='datetime64') ==
            np.datetime64('2014-12-27T00:00:00.000000'))
    assert (dtt.get_date_time_from_metadata(md1, formatting='datetime') ==
            dt1)

    assert (dtt.get_date_time_from_metadata(md2) ==
            '2124-03-25T10:04:48-05:00')
    assert (dtt.get_date_time_from_metadata(md2, formatting='datetime') ==
            dt2)
    assert (dtt.get_date_time_from_metadata(md2, formatting='datetime64') ==
            np.datetime64('2124-03-25T10:04:48'))

    assert (dtt.get_date_time_from_metadata(md3) ==
            '2016-07-12T22:57:32')
    assert (dtt.get_date_time_from_metadata(md3, formatting='datetime') ==
            dt3)
    assert (dtt.get_date_time_from_metadata(md3, formatting='datetime64') ==
            np.datetime64('2016-07-12T22:57:32.000000'))

    assert (dtt.get_date_time_from_metadata(DictionaryTreeBrowser({'General': {}})) ==
            None)
    assert (dtt.get_date_time_from_metadata(DictionaryTreeBrowser({'General': {'date': '2016-07-12'}})) ==
            '2016-07-12')
    assert (dtt.get_date_time_from_metadata(DictionaryTreeBrowser({'General': {'time': '12:00'}})) ==
            '12:00:00')
    assert (dtt.get_date_time_from_metadata(DictionaryTreeBrowser({'General': {'time': '12:00',
                                                                               'time_zone': 'CET'}})) ==
            '12:00:00')


def _defaut_metadata():
    md = DictionaryTreeBrowser({'General': {'date': '2016-01-01',
                                            'time': '00:00:00',
                                            'time_zone': 'GMT'}})
    return md


def test_update_date_time_in_metadata():
    md = DictionaryTreeBrowser({'General': {}})
    # in case of iso, the exact time is lost, only the time offset is kept
    md11 = dtt.update_date_time_in_metadata(iso1, md.deepcopy())
    assert_deep_almost_equal(md11.General.date, md1.General.date)
    assert_deep_almost_equal(md11.General.time, md1.General.time)
    assert_deep_almost_equal(md11.General.time_zone, 'UTC')

    md12 = dtt.update_date_time_in_metadata(dt1, md.deepcopy())
    assert_deep_almost_equal(md12.General.date, md1.General.date)
    assert_deep_almost_equal(md12.General.time, md1.General.time)
    import locale
    if locale.getlocale()[0] in ['en_GB', 'en_US']:
        assert md12.General.time_zone in ('UTC', 'Coordinated Universal Time')

    md13 = dtt.update_date_time_in_metadata(iso2, md.deepcopy())
    assert_deep_almost_equal(md13.General.date, md2.General.date)
    assert_deep_almost_equal(md13.General.time, md2.General.time)
    assert_deep_almost_equal(md13.General.time_zone, '-05:00')
    assert_deep_almost_equal(dtt.update_date_time_in_metadata(dt2, md.deepcopy()).as_dictionary(),
                             md2.as_dictionary())

    assert_deep_almost_equal(dtt.update_date_time_in_metadata(iso3, md.deepcopy()).as_dictionary(),
                             md3.as_dictionary())
    assert_deep_almost_equal(dtt.update_date_time_in_metadata(dt3, md.deepcopy()).as_dictionary(),
                             md3.as_dictionary())


def test_serial_date_to_ISO_format():
    iso_1 = dtt.serial_date_to_ISO_format(serial1)
    dt1_local = dt1.astimezone(tz.tzlocal())
    assert iso_1[0] == dt1_local.date().isoformat()
    assert iso_1[1] == dt1_local.time().isoformat()
    assert iso_1[2] == dt1_local.tzname()

    iso_2 = dtt.serial_date_to_ISO_format(serial2)
    dt2_local = dt2.astimezone(tz.tzlocal())
    assert iso_2[0] == dt2_local.date().isoformat()
    # The below line will/can fail due to accuracy loss when converting to serial date:
    # We therefore truncate milli/micro seconds
    assert iso_2[1][:8] == dt2_local.time().isoformat()
    assert iso_2[2] == dt2_local.tzname()

    iso_3 = dtt.serial_date_to_ISO_format(serial3)
    dt3_aware = dt3.replace(tzinfo=tz.tzutc())
    dt3_local = dt3_aware.astimezone(tz.tzlocal())
    assert iso_3[0] == dt3_local.date().isoformat()
    assert iso_3[1] == dt3_local.time().isoformat()
    assert iso_3[2] == dt3_local.tzname()


def test_ISO_format_to_serial_date():
    res1 = dtt.ISO_format_to_serial_date(
        dt1.date().isoformat(), dt1.time().isoformat(), timezone=dt1.tzname())
    np.testing.assert_allclose(res1, serial1, atol=1E-5)
    dt = dt2.astimezone(tz.tzlocal())
    res2 = dtt.ISO_format_to_serial_date(
        dt.date().isoformat(), dt.time().isoformat(), timezone=dt.tzname())
    np.testing.assert_allclose(res2, serial2, atol=1E-5)
    res3 = dtt.ISO_format_to_serial_date(
        dt3.date().isoformat(), dt3.time().isoformat(), timezone=dt3.tzname())
    np.testing.assert_allclose(res3, serial3, atol=1E-5)


def test_datetime_to_serial_date():
    np.testing.assert_allclose(dtt.datetime_to_serial_date(dt1), serial1, atol=1E-5)
    np.testing.assert_allclose(dtt.datetime_to_serial_date(dt2), serial2, atol=1E-5)
    np.testing.assert_allclose(dtt.datetime_to_serial_date(dt3), serial3, atol=1E-5)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import time
import logging
from multiprocessing import Manager
import numpy as np
from dask.array import Array as dar

from hyperspy.utils.parallel_pool import ParallelPool
from hyperspy.samfire_utils.samfire_worker import create_worker

_logger = logging.getLogger(__name__)


def _walk_compute(athing):
    if isinstance(athing, dict):
        this = {}
        for key, val in athing.items():
            if isinstance(key, dar):
                raise ValueError('Dask arrays should not be used as keys')
            value = _walk_compute(val)
            this[key] = value
        return this
    elif isinstance(athing, list):
        return [_walk_compute(val) for val in athing]
    elif isinstance(athing, tuple):
        return tuple(_walk_compute(val) for val in athing)
    elif isinstance(athing, dar):
        _logger.debug('found a dask array!')
        return athing.compute()
    else:
        return athing


class SamfirePool(ParallelPool):
    """ Creates and manages a pool of SAMFire workers. For based on
    ParallelPool - either creates processes using multiprocessing, or connects
    and sets up ipyparallel load_balanced_view.

    Ipyparallel is managed directly, but multiprocessing pool is managed via
    three of queues:

    * Shared by all (master and workers) for distributing "load-balanced"
      work.
    * Shared by all (master and workers) for sending results back to the
      master
    * Individual queues from master to each worker. For setting up and
      addressing individual workers in general. This one is checked with
      higher priority in workers.

    Methods
    -------
    prepare_workers
        Given SAMFire object, populates the workers with the required
        information. In case of multiprocessing, starts worker listening to the
        queues.
    update_parameters
        Updates various worker parameters
    ping_workers
        Pings all workers. Stores the one-way trip time and the process_id
        (pid) of each worker if available
    add_jobs
        Adds the requested number of jobs to the queue
    parse
        Parses the messages, returned from the workers.
    colect_results
        Collects all currently available results and parses them
    run
        Runs the full procedure until no more pixels are left to run in the
        SAMFire
    stop
        Stops the pool, (for ipyparallel) clears the memory
    setup
        Sets up the ipyparallel or multiprocessing pool (collects to the
        client or creates the pool)
    sleep
        Sleeps for the specified time, by default timestep

    Attributes
    ----------
    has_pool : bool
        Boolean if the pool is available and active
    pool : {ipyparallel.load_balanced_view, multiprocessing.Pool}
        The pool object
    ipython_kwargs : dict
        The dictionary with Ipyparallel connection arguments.
    timeout : float
        Timeout for either pool when waiting for results
    num_workers : int
        The number of workers actually created (may be less than requested, but
        can't be more)
    timestep : float
        The timestep between "ticks" that the result queues are checked. Higher
        timestep means less frequent checking, which may reduce CPU load for
        difficult fits that take a long time to finish.
    ping : dict
        If recorded, stores one-way trip time of each worker
    pid : dict
        If available, stores the process-id of each worker
    """

    def __init__(self, **kwargs):
        """Creates a ParallelPool with additional methods for SAMFire. All
        arguments are passed to ParallelPool"""
        super(SamfirePool, self).__init__(**kwargs)
        self.samf = None
        self.ping = {}
        self.pid = {}
        self.workers = {}
        self.rworker = None
        self.result_queue = None
        self.shared_queue = None
        self._last_time = 0
        self.results = []

    def _timestep_set(self, value):
        value = np.abs(value)
        self._timestep = value
        if self.has_pool and self.is_multiprocessing:
            for this_queue in self.workers.values():
                this_queue.put(('change_timestep', (value,)))

    def prepare_workers(self, samfire):
        """Prepares the workers for work, in case of multiprocessing starts
        listening

        Parameters
        ----------
        samfire : samfire
            the SAMFire object that will be using the pool
        """
        _logger.debug('starting prepare_workers')
        self.samf = samfire
        mall = samfire.model
        model = mall.inav[mall.axes_manager.indices]
        if model.signal.metadata.has_item('Signal.Noise_properties.variance'):
            var = model.signal.metadata.Signal.Noise_properties.variance
            if var._lazy:
                var.compute()
        model.store('z')
        m_dict = model.signal._to_dictionary(False)
        m_dict['models'] = model.signal.models._models.as_dictionary()

        m_dict = _walk_compute(m_dict)

        optional_names = {mall[c].name for c in samfire.optional_components}

        if self.is_ipyparallel:
            from ipyparallel import Reference as ipp_Reference
            _logger.debug('preparing ipyparallel workers')
            direct_view = self.pool.client[:self.num_workers]
            direct_view.block = True
            direct_view.execute("from hyperspy.samfire_utils.samfire_worker"
                                " import create_worker")
            direct_view.scatter('identity', range(self.num_workers),
                                flatten=True)
            direct_view.execute('worker = create_worker(identity)')
            self.rworker = ipp_Reference('worker')
            direct_view.apply(lambda worker, m_dict:
                              worker.create_model(m_dict, 'z'), self.rworker,
                              m_dict)
            direct_view.apply(lambda worker, ts: worker.setup_test(ts),
                              self.rworker, samfire.metadata._gt_dump)
            direct_view.apply(lambda worker, on: worker.set_optional_names(on),
                              self.rworker, optional_names)

        if self.is_multiprocessing:
            _logger.debug('preparing multiprocessing workers')
            manager = Manager()
            self.shared_queue = manager.Queue()
            self.result_queue = manager.Queue()
            for i in range(self.num_workers):
                this_queue = manager.Queue()
                self.workers[i] = this_queue
                this_queue.put(('setup_test', (samfire.metadata._gt_dump,)))
                this_queue.put(('create_model', (m_dict, 'z')))
                this_queue.put(('set_optional_names', (optional_names,)))
                self.pool.apply_async(create_worker, args=(i, this_queue,
                                                           self.shared_queue,
                                                           self.result_queue))

    def update_parameters(self):
        """Updates various worker parameters.

        Currently updates:
            - Optional components (that can be switched off by the worker)
            - Parameter boundaries
            - Goodness test"""
        samfire = self.samf
        optional_names = {samfire.model[c].name for c in
                          samfire.optional_components}
        boundaries = tuple(tuple((par.bmin, par.bmax) for par in
                                 comp.parameters) for comp in self.samf.model)
        if self.is_multiprocessing:
            for this_queue in self.workers.values():
                this_queue.put(('set_optional_names', (optional_names,)))
                this_queue.put(('setup_test', (samfire.metadata._gt_dump,)))
                this_queue.put(('set_parameter_boundaries', (boundaries,)))
        elif self.is_ipyparallel:
            direct_view = self.pool.client[:self.num_workers]
            direct_view.block = True
            direct_view.apply(lambda worker, on: worker.set_optional_names(on),
                              self.rworker, optional_names)
            direct_view.apply(lambda worker, ts: worker.setup_test(ts),
                              self.rworker, samfire.metadata._gt_dump)
            direct_view.apply(lambda worker, ts:
                              worker.set_parameter_boundaries(ts),
                              self.rworker, boundaries)

    def ping_workers(self, timeout=None):
        """Pings the workers and records one-way trip time and (if available)
        pid of the worker.

        Parameters
        ----------
        timeout: {None, flaot}
            the time to wait when collecting results after sending out the
            ping. If None, the default timeout is used
        """
        if self.samf is None:
            _logger.error('Have to add samfire to the pool first')
        else:
            if self.is_multiprocessing:
                for _id, this_queue in self.workers.items():
                    this_queue.put('ping')
                    self.ping[_id] = time.time()
            elif self.is_ipyparallel:
                for i in range(self.num_workers):
                    direct_view = self.pool.client[i]
                    self.results.append((direct_view.apply_async(lambda worker:
                                                                 worker.ping(),
                                                                 self.rworker),
                                         i))
                    self.ping[i] = time.time()
        time.sleep(0.5)
        self.collect_results(timeout)

    def __len__(self):
        if self.is_ipyparallel:
            return self.pool.client.queue_status()['unassigned']
        elif self.is_multiprocessing:
            return self.shared_queue.qsize()

    def add_jobs(self, needed_number=None):
        """Adds jobs to the job queue that is consumed by the workers.

        Parameters
        ----------
        needed_number: {None, int}
            The number of jobs to add. If None (default), adds `need_pixels`
        """
        def test_func(worker, ind, value_dict):
            return worker.run_pixel(ind, value_dict)
        if needed_number is None:
            needed_number = self.need_pixels
        for ind, value_dict in self.samf.generate_values(needed_number):
            if self.is_multiprocessing:
                self.shared_queue.put(('run_pixel', (ind, value_dict)))
            elif self.is_ipyparallel:
                self.results.append((self.pool.apply_async(test_func,
                                                           self.rworker, ind,
                                                           value_dict), ind))

    def parse(self, value):
        """Parses the value, returned from the workers.

        Parameters
        ----------
        value: tuple of the form (keyword, the_rest)
            Keyword currently can be one of ['pong', 'Error', 'result']. For
            each of the keywords, "the_rest" is a tuple of different elements,
            but generally the first one is always the worker_id that the result
            came from. In particular:

            * ('pong', (worker_id, pid, pong_time, optional_message_str))
            * ('Error', (worker_id, error_message_string))
            * ('result', (worker_id, pixel_index, result_dict,
              bool_if_result_converged))
        """
        if value is None:
            keyword = 'Failed'
            _logger.debug('Got None')
        else:
            keyword, the_rest = value
        samf = self.samf
        if keyword == 'pong':
            _id, pid, pong_time, message = the_rest
            self.ping[_id] = pong_time - self.ping[_id]
            self.pid[_id] = pid
            _logger.info('pong worker %s with time %g and message'
                         '"%s"' % (str(_id), self.ping[_id], message))
        elif keyword == 'Error':
            _id, err_message = the_rest
            _logger.error('Error in worker %s\n%s' % (str(_id), err_message))
        elif keyword == 'result':
            _id, ind, result, isgood = the_rest
            _logger.debug('Got result from pixel {} and it is good:'
                          '{}'.format(ind, isgood))
            if ind in samf.running_pixels:
                samf.running_pixels.remove(ind)
                samf.update(ind, result, isgood)
                samf.plot(on_count=True)
                samf.backup()
                samf.log(ind, isgood, samf.count, _id)
        else:
            _logger.error('Unusual return from some worker. The value '
                          'is:\n%s' % str(value))

    def collect_results(self, timeout=None):
        """Collects and parses all results, currently not processed due to
        being in the queue.

        Parameters
        ----------
        timeout: {None, flaot}
            the time to wait when collecting results. If None, the default
            timeout is used

        """
        if timeout is None:
            timeout = self.timeout
        found_something = False
        if self.is_ipyparallel:
            # for res, ind in reversed(self.results):
            for res, ind in self.results:
                if res.ready():
                    try:
                        result = res.get(timeout=timeout)
                    except TimeoutError:
                        _logger.info('Ind {} failed to come back in {} '
                                     'seconds. Assuming failed'.format(
                                         ind, timeout))
                        result = ('result', (-1, ind, None, False))
                    self.parse(result)
                    self.results.remove((res, ind))
                    found_something = True
                else:
                    pass
        elif self.is_multiprocessing:
            while not self.result_queue.empty():
                try:
                    result = self.result_queue.get(block=True,
                                                   timeout=timeout)
                    self.parse(result)
                    found_something = True
                except TimeoutError:
                    _logger.info('Some ind failed to come back in {} '
                                 'seconds.'.format(self.timeout))
        return found_something

    @property
    def need_pixels(self):
        """Returns the number of pixels that should be added to the processing
        queue. At most is equal to the number of workers.
        """
        return min(self.samf.pixels_done * self.samf.metadata.marker.ndim,
                   self.num_workers - len(self))

    @property
    def _not_too_long(self):
        """Returns bool if it has been too long after receiving the last
        result, probably meaning some of the workers timed out or hung.
        """
        if not hasattr(self, '_last_time') or not isinstance(self._last_time,
                                                             float):
            self._last_time = time.time()
        return (time.time() - self._last_time) <= self.timeout

    def run(self):
        """Runs the full process of adding jobs to the processing queue,
        listening to the results and updating SAMFire as needed. Stops when
        timed out or no pixels are left to run.
        """
        while self._not_too_long and (self.samf.pixels_left or
                                      len(self.samf.running_pixels)):
            # bool if got something
            new_result = self.collect_results()
            need_number = self.need_pixels

            if need_number > 0:
                self.add_jobs(need_number)
            if not need_number or not new_result:
                # did not spend much time, since no new results or added pixels
                self.sleep()
            else:
                self._last_time = time.time()

    def stop(self):
        """Stops the appropriate pool and (if ipyparallel) clears the memory
        and history.
        """
        if self.is_multiprocessing:
            for queue in self.workers.values():
                queue.put('stop_listening')
            self.pool.close()
            self.pool.terminate()
            self.pool.join()
        elif self.is_ipyparallel:
            self.pool.client.clear()



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from hyperspy.decorators import lazifyTestClass
from hyperspy.signals import Signal1D


def _test_default_navigation_signal_operations_over_many_axes(self, op):
    s = getattr(self.signal, op)()
    ar = getattr(self.data, op)(axis=(0, 1))
    np.testing.assert_array_equal(ar, s.data)
    assert s.data.ndim == 1
    assert s.axes_manager.signal_dimension == 1
    assert s.axes_manager.navigation_dimension == 0


@lazifyTestClass
class Test3D:
    def setup_method(self, method):
        self.signal = Signal1D(np.arange(2 * 4 * 6).reshape(2, 4, 6))
        self.signal.axes_manager[0].name = "x"
        self.signal.axes_manager[1].name = "y"
        self.signal.axes_manager[2].name = "E"
        self.signal.axes_manager[0].scale = 0.5
        self.data = self.signal.data.copy()

    def test_indexmin(self):
        s = self.signal.indexmin("E")
        ar = self.data.argmin(2)
        np.testing.assert_array_equal(ar, s.data)
        assert s.data.ndim == 2
        assert s.axes_manager.signal_dimension == 0
        assert s.axes_manager.navigation_dimension == 2

    def test_indexmax(self):
        s = self.signal.indexmax("E")
        ar = self.data.argmax(2)
        np.testing.assert_array_equal(ar, s.data)
        assert s.data.ndim == 2
        assert s.axes_manager.signal_dimension == 0
        assert s.axes_manager.navigation_dimension == 2

    def test_valuemin(self):
        s = self.signal.valuemin("x")
        ar = self.signal.axes_manager["x"].index2value(self.data.argmin(1))
        np.testing.assert_array_equal(ar, s.data)
        assert s.data.ndim == 2
        assert s.axes_manager.signal_dimension == 1
        assert s.axes_manager.navigation_dimension == 1

    def test_valuemax(self):
        s = self.signal.valuemax("x")
        ar = self.signal.axes_manager["x"].index2value(self.data.argmax(1))
        np.testing.assert_array_equal(ar, s.data)
        assert s.data.ndim == 2
        assert s.axes_manager.signal_dimension == 1
        assert s.axes_manager.navigation_dimension == 1

    def test_default_navigation_sum(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "sum")

    def test_default_navigation_max(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "max")

    def test_default_navigation_min(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "min")

    def test_default_navigation_mean(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "mean")

    def test_default_navigation_std(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "std")

    def test_default_navigation_var(self):
        _test_default_navigation_signal_operations_over_many_axes(self, "var")

    def test_rebin(self):
        self.signal.estimate_poissonian_noise_variance()
        new_s = self.signal.rebin(scale=(2, 2, 1))
        var = new_s.metadata.Signal.Noise_properties.variance
        assert new_s.data.shape == (1, 2, 6)
        assert var.data.shape == (1, 2, 6)
        from hyperspy.misc.array_tools import rebin

        np.testing.assert_array_equal(
            rebin(self.signal.data, scale=(2, 2, 1)), var.data
        )
        np.testing.assert_array_equal(
            rebin(self.signal.data, scale=(2, 2, 1)), new_s.data
        )
        if self.signal._lazy:
            new_s = self.signal.rebin(scale=(2, 2, 1), rechunk=False)
            np.testing.assert_array_equal(
                rebin(self.signal.data, scale=(2, 2, 1)), var.data
            )
            np.testing.assert_array_equal(
                rebin(self.signal.data, scale=(2, 2, 1)), new_s.data
            )

    def test_rebin_no_variance(self):
        new_s = self.signal.rebin(scale=(2, 2, 1))
        with pytest.raises(AttributeError):
            _ = new_s.metadata.Signal.Noise_properties

    def test_rebin_const_variance(self):
        self.signal.metadata.set_item("Signal.Noise_properties.variance", 0.3)
        new_s = self.signal.rebin(scale=(2, 2, 1))
        assert new_s.metadata.Signal.Noise_properties.variance == 0.3

    def test_rebin_dtype(self):
        s = Signal1D(np.arange(1000).reshape(10, 10, 10))
        s.change_dtype(np.uint8)
        s2 = s.rebin(scale=(3, 3, 1), crop=False)
        assert s.sum() == s2.sum()

    def test_swap_axes_simple(self):
        s = self.signal
        assert s.swap_axes(0, 1).data.shape == (4, 2, 6)
        assert s.swap_axes(0, 2).axes_manager.shape == (6, 2, 4)
        if not s._lazy:
            assert not s.swap_axes(0, 2).data.flags["C_CONTIGUOUS"]
            assert s.swap_axes(0, 2, optimize=True).data.flags["C_CONTIGUOUS"]
        else:
            cks = s.data.chunks
            assert s.swap_axes(0, 1).data.chunks == (cks[1], cks[0], cks[2])
            # This data shape does not require rechunking
            assert s.swap_axes(0, 1, optimize=True).data.chunks == (
                cks[1],
                cks[0],
                cks[2],
            )

    def test_swap_axes_iteration(self):
        s = self.signal
        s = s.swap_axes(0, 2)
        assert s.axes_manager._getitem_tuple[:2] == (0, 0)
        s.axes_manager.indices = (2, 1)
        assert s.axes_manager._getitem_tuple[:2] == (1, 2)

    def test_get_navigation_signal_nav_dim0(self):
        s = self.signal
        s = s.transpose(signal_axes=3)
        ns = s._get_navigation_signal()
        assert ns.axes_manager.signal_dimension == 1
        assert ns.axes_manager.signal_size == 1
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_navigation_signal_nav_dim1(self):
        s = self.signal
        s = s.transpose(signal_axes=2)
        ns = s._get_navigation_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.navigation_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_navigation_signal_nav_dim2(self):
        s = self.signal
        s = s.transpose(signal_axes=1)
        ns = s._get_navigation_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.navigation_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_navigation_signal_nav_dim3(self):
        s = self.signal
        s = s.transpose(signal_axes=0)
        ns = s._get_navigation_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.navigation_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_navigation_signal_wrong_data_shape(self):
        s = self.signal
        s = s.transpose(signal_axes=1)
        with pytest.raises(ValueError):
            s._get_navigation_signal(data=np.zeros((3, 2)))

    def test_get_navigation_signal_wrong_data_shape_dim0(self):
        s = self.signal
        s = s.transpose(signal_axes=3)
        with pytest.raises(ValueError):
            s._get_navigation_signal(data=np.asarray(0))

    def test_get_navigation_signal_given_data(self):
        s = self.signal
        s = s.transpose(signal_axes=1)
        data = np.zeros(s.axes_manager._navigation_shape_in_array)
        ns = s._get_navigation_signal(data=data)
        assert ns.data is data

    def test_get_signal_signal_nav_dim0(self):
        s = self.signal
        s = s.transpose(signal_axes=0)
        ns = s._get_signal_signal()
        assert ns.axes_manager.navigation_dimension == 0
        assert ns.axes_manager.navigation_size == 0
        assert ns.axes_manager.signal_dimension == 1

    def test_get_signal_signal_nav_dim1(self):
        s = self.signal
        s = s.transpose(signal_axes=1)
        ns = s._get_signal_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.signal_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_signal_signal_nav_dim2(self):
        s = self.signal
        s = s.transpose(signal_axes=2)
        s._assign_subclass()
        ns = s._get_signal_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.signal_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_signal_signal_nav_dim3(self):
        s = self.signal
        s = s.transpose(signal_axes=3)
        s._assign_subclass()
        ns = s._get_signal_signal()
        assert ns.axes_manager.signal_shape == s.axes_manager.signal_shape
        assert ns.axes_manager.navigation_dimension == 0

    def test_get_signal_signal_wrong_data_shape(self):
        s = self.signal
        s = s.transpose(signal_axes=1)
        with pytest.raises(ValueError):
            s._get_signal_signal(data=np.zeros((3, 2)))

    def test_get_signal_signal_wrong_data_shape_dim0(self):
        s = self.signal
        s = s.transpose(signal_axes=0)
        with pytest.raises(ValueError):
            s._get_signal_signal(data=np.asarray(0))

    def test_get_signal_signal_given_data(self):
        s = self.signal
        s = s.transpose(signal_axes=2)
        data = np.zeros(s.axes_manager._signal_shape_in_array)
        ns = s._get_signal_signal(data=data)
        assert ns.data is data

    def test_get_navigation_signal_dtype(self):
        s = self.signal
        assert s._get_navigation_signal().data.dtype.name == s.data.dtype.name

    def test_get_signal_signal_dtype(self):
        s = self.signal
        assert s._get_signal_signal().data.dtype.name == s.data.dtype.name

    def test_get_navigation_signal_given_dtype(self):
        s = self.signal
        assert s._get_navigation_signal(dtype="bool").data.dtype.name == "bool"

    def test_get_signal_signal_given_dtype(self):
        s = self.signal
        assert s._get_signal_signal(dtype="bool").data.dtype.name == "bool"
